from collections import deque


class ResultsCollector:
    def __init__(self, len):
        self._data = deque(maxlen=len)
        self._last_insertion_position = 0

    def append(self, data):
        self._data.append(data)

        if self._last_insertion_position != self._get_size():
            self._last_insertion_position += 1

    def _get_size(self):
        return len(self._data)

    def get_data(self):
        return list(self._data)[:self._last_insertion_position]

    def clear(self):
        self._data.clear()




from ate_master_app.parameter_parser.xml_parameter_parser import XmlParameterParser
from ate_master_app.parameter_parser.filesystem_data_source import FileSystemDataSource
from ate_master_app.parameter_parser.static_data_source import StaticDataSource


def CreateParser(jobformat: str):
    if(jobformat == 'xml.semi-ate'):
        return XmlParameterParser()

    return None


def CreateDataSource(jobname: str, configuration: dict, parser, logger):
    jobdatasource = configuration['jobsource']

    if(jobdatasource == 'filesystem'):
        return FileSystemDataSource(jobname, configuration, parser, logger)
    if(jobdatasource == "static"):
        return StaticDataSource()

    return None



""" master application """
# from common.application import Application
from ate_master_app.master_application import MasterApplication, assert_valid_system_mimetypes_config
from ate_apps_common import configuration_reader
from argparse import ArgumentParser


DEFAULT_LOGFILE_NAME = "master_app.log"
DEFAULT_CONFIGFILE_PATH = "./master_config_file.json"


def parse_config_dict_from_command_line():
    parser = ArgumentParser()
    parser.add_argument('-d', '--device-id', help='device id', type=str)
    parser.add_argument('-s', '--sites', help='comma delimited list of site-ids', type=lambda s: [item.strip() for item in s.split(',')])
    parser.add_argument('-f', '--file', help='configuration file name', type=lambda s: [item.strip() for item in s.split(',')])
    args = parser.parse_args()

    config = {}
    if args.device_id is not None:
        config['device_id'] = args.device_id
    if args.sites is not None:
        config['sites'] = args.sites

    if args.file is not None:
        config['file'] = args.file

    return config


def launch_master(*, config_file_path=None, user_config_dict=None):
    assert_valid_system_mimetypes_config()

    cfg = configuration_reader.ConfigReader(config_file_path)
    configuration = cfg.get_configuration_ex(user_config_dict=user_config_dict)

    f = MasterApplication(configuration)
    f.run()


def main():
    user_config = parse_config_dict_from_command_line()
    launch_master(config_file_path=DEFAULT_CONFIGFILE_PATH,
                  user_config_dict=user_config)

if __name__ == '__main__':
    main()



from transitions.extensions import HierarchicalMachine as Machine
from typing import List, Optional

from ate_master_app.statemachines.TestsExecutionMachine import TestsExecutionMachine


class TestingSiteMachine(Machine):
    states = ['inprogress',  # waiting
              'completed',
              'strategy_info_confirmed',
              {'name': 'testing', 'children': TestsExecutionMachine()},
              'error']

    def __init__(self, model):
        super().__init__(model=model, states=self.states, initial='testing', send_event=True)

        self.add_transition('testrequest_received',     'testing_idle',                             'testing_waiting_for_release'),                                              # noqa: E241
        self.add_transition('testrequest_received',     ['testing_busy', 'testing_waiting_for_resource'],             'testing_waiting_for_release'),                                              # noqa: E241
        self.add_transition('testrequest_released',     'testing_waiting_for_release',              'testing_busy'),                                                       # noqa: E241

        self.add_transition('resource_requested',       'testing_busy',                             'testing_waiting_for_resource',     before='set_requested_resource')        # noqa: E241
        self.add_transition('resource_ready',           'testing_waiting_for_resource',             'testing_busy',                     before='clear_requested_resource')      # noqa: E241

        self.add_transition('on_unload',                '*',                                        '=',                       after='clear_execution_strategy'),                             # noqa: E241

        self.add_transition('on_error',                 '*',                                        'error'),                                            # noqa: E241
        self.add_transition('reset',                    '*',                                        'inprogress',                       before=['clear_testresult', 'clear_execution_strategy']),              # noqa: E241


class TestingSiteModel:
    site_id: str
    resource_request: Optional[dict]
    testresult: Optional[dict]
    execution_strategy: Optional[List]

    def __init__(self, site_id: str):
        self.site_id = site_id
        self.resource_request = None
        self.testresult = None
        self.execution_strategy = None

    def set_requested_resource(self, event):
        self.resource_request = event.kwargs['resource_request']

    def clear_requested_resource(self, event):
        self.resource_request = None

    def set_testresult(self, event):
        self.testresult = event.kwargs['testresult']

    def clear_testresult(self, event):
        self.testresult = None

    def set_execution_strategy(self, strategy: List[List[str]]):
        self.execution_strategy = strategy

    def clear_execution_strategy(self, _):
        self.execution_strategy = None



import json
import asyncio


class PeripheralController:
    def __init__(self, mqtt_client, device_id):
        self.mqtt_client = mqtt_client
        self.device_id = device_id
        self.ioctl_response_event = asyncio.Event()

    async def device_io_control(self, peripheral_name, ioctl_code, parameters):
        message = {"type": "io-control-request",
                   "ioctl_name": ioctl_code,
                   "parameters": parameters
                   }

        request_topic = f"ate/{self.device_id}/{peripheral_name}/io-control"
        self.ioctl_response_event.clear()
        return await self.mqtt_client.publish_with_response(request_topic, json.dumps(message))

    def on_response_received(self, response):
        self.received_response = response
        self.ioctl_response_event.set()



from ate_common.logger import LogLevel
from aiohttp import web, WSMsgType, WSCloseCode
from pathlib import Path
import json
import time
import os
from enum import Enum
import uuid


class MessageTypes(Enum):
    TestResult = 'testresult'
    Status = 'status'
    UserSettings = 'usersettings'
    TestResults = 'testresults'
    Logs = 'logs'
    Logfile = 'logfile'
    Yield = 'yield'
    LotData = 'lotdata'
    BinTable = 'bintable'
    Configuration = 'masterconfiguration'

    def __call__(self):
        return self.value


class WebsocketCommunicationHandler:
    def __init__(self, app):
        self._app = app
        self._websockets = set()
        self._log = app['master_app'].log

    def get_current_master_state(self):
        master_app = self._app['master_app']
        return master_app.external_state

    def get_broker_from_master_config(self):
        master_app = self._app['master_app']
        return (master_app.configuration.get('broker_host'),
                master_app.configuration.get('broker_port'))

    def get_mqtt_handler(self):
        return self._app['mqtt_handler']

    async def send_message_to_all(self, data):
        self._discard_ws_connection_if_needed()
        for ws in self._websockets:
            await ws.send_json(data)

    async def send_message_to_client(self, client_id, data):
        self._discard_ws_connection_if_needed()
        for ws in self._websockets:
            if ws.get_connection_id() == client_id:
                await ws.send_json(data)

    async def send_status_to_all(self, state, description):
        status_message = self._create_status_message(state, description)
        await self.send_message_to_all(status_message)

    async def send_configuration(self, configuration, client_ws):
        configuration_message = self._create_configuration_message(configuration)
        await self.send_message_to_client(client_ws.connection_id, configuration_message)

    async def send_testresults_to_all(self, site_test_result):
        await self.send_message_to_all(site_test_result)

    async def send_testresults_to_client(self, testresults, connection_id):
        testresults_message = self._create_testresults_message(testresults)
        await self.send_message_to_client(connection_id, testresults_message)

    async def send_logs_to_all(self, logs):
        logs_message = self._create_logs_message(logs)
        await self.send_message_to_all(logs_message)

    async def send_logs(self, logs, connection_id):
        logs_message = self._create_logs_message(logs)
        await self.send_message_to_client(connection_id, logs_message)

    async def send_logfile(self, logfile_info, connection_id):
        logs_message = self._create_logfile_message(logfile_info)
        await self.send_message_to_client(connection_id, logs_message)

    async def send_user_settings(self, usersettings):
        testresults_message = self._create_usersettings_message(usersettings)
        await self.send_message_to_all(testresults_message)

    async def send_yields(self, yields):
        yields_message = self._create_yield_message(yields)
        await self.send_message_to_all(yields_message)

    async def send_lotdata(self, lotdata):
        lotdata_message = self._create_lotdata_message(lotdata)
        await self.send_message_to_all(lotdata_message)

    async def send_bin_table(self, bin_table):
        yields_message = self._create_bin_table_message(bin_table)
        await self.send_message_to_all(yields_message)

    async def _send_connection_id_to_ws(self, ws, connection_id):
        connection_id_message = self._create_connection_id_message(connection_id)
        await ws.send_json(connection_id_message)

    def _create_connection_id_message(self, connection_id):
        payload = {"connectionid": connection_id}
        return self._create_message("connectionid", payload)

    @staticmethod
    def _create_message(type, payload):
        return {'type': type, 'payload': payload}

    def _create_testresults_message(self, testresults):
        return self._create_message(MessageTypes.TestResults(), testresults)

    def _create_logs_message(self, logs):
        return self._create_message(MessageTypes.Logs(), logs)

    def _create_logfile_message(self, logfile_info):
        return self._create_message(MessageTypes.Logfile(), logfile_info)

    def _create_usersettings_message(self, user_settings):
        return self._create_message(MessageTypes.UserSettings(), user_settings)

    def _create_yield_message(self, yields):
        return self._create_message(MessageTypes.Yield(), yields)

    def _create_bin_table_message(self, bin_table: list):
        return self._create_message(MessageTypes.BinTable(), bin_table)

    def _create_lotdata_message(self, lotdata):
        return self._create_message(MessageTypes.LotData(), lotdata)

    def _create_status_message(self, state, error_message):
        payload = {'device_id': self._app['master_app'].device_id,
                   'systemTime': time.strftime("%b %d %Y %H:%M:%S"),
                   'sites': self._app['master_app'].configuredSites,
                   'state': state,
                   'error_message': error_message,
                   'env': self._app['master_app'].env,
                   'lot_number': self._app['master_app'].loaded_lot_number}

        return self._create_message(MessageTypes.Status(), payload)

    def _create_configuration_message(self, configuration):
        payload = {'broker_host': configuration['broker_host'],
                   'broker_port': configuration['broker_port'],
                   'device_id': self._app['master_app'].device_id,
                   'sites': self._app['master_app'].configuredSites,
                   'handler': configuration['Handler'],
                   'environment': configuration['environment'],
                   'jobsource': configuration['jobsource'],
                   'jobformat': configuration['jobformat'],
                   }

        return self._create_message(MessageTypes.Configuration(), payload)

    async def close_all(self):
        for ws in set(self._websockets):
            await ws.close(code=WSCloseCode.GOING_AWAY,
                           message='Server shutdown')

    async def receive(self, request):
        ws = web.WebSocketResponse(heartbeat=1.0)
        await ws.prepare(request)

        connection_id = str(uuid.uuid1())
        ws_connection = WebSocketConnection(ws, connection_id)
        self._websockets.add(ws_connection)

        await self._send_connection_id_to_ws(ws, connection_id)

        # master should propagate the available settings each time a page reloaded
        # or new websocket connection is required
        self.handle_new_connection(connection_id)
        await self.send_configuration(self._app['master_app'].configuration, ws_connection)

        self._log.log_message(LogLevel.Debug(), 'websocket connection opened.')

        try:
            async for msg in ws:
                if msg.type == WSMsgType.TEXT:
                    self.handle_client_message(msg.data)
                elif msg.type == WSMsgType.ERROR:
                    self._log.log_message(LogLevel.Error(), f'ws connection closed with exception: {ws.exception()}')
        finally:
            pass

        self._discard_ws_connection_if_needed()

    def _discard_ws_connection_if_needed(self):
        self._websockets = set([ws for ws in self._websockets if ws.is_alive()])

    def handle_new_connection(self, connection_id):
        self._app['master_app'].on_new_connection(connection_id)

    def handle_client_message(self, message_data: str):
        json_data = json.loads(message_data)
        self._log.log_message(LogLevel.Debug(), f'Server received message {json_data}')

        if json_data.get('type') == 'cmd':
            self._app["master_app"].dispatch_command(json_data)

    def is_ws_connection_established(self):
        return len(self._websockets) > 0


async def index_handler(request):
    static_file_path = request.app['static_file_path']
    return web.FileResponse(os.path.join(static_file_path, 'index.html'))


async def webservice_init(app):
    static_file_path = app['static_file_path']
    ws_comm_handler = WebsocketCommunicationHandler(app)
    if Path(static_file_path, 'index.html').exists() == False:
        async def index(request: web.Request) -> web.Response:
            text = '''
            Please configure the JSON-key "webui_root_path" in file "master_config_file.json" in such a way
            that it points to a web-root folder, i.e. a folder containing some index.html and restart the
            master application.        
            '''
            return web.Response(text=text)
        global index_handler
        index_handler = index
    app.add_routes([web.get('/', index_handler),
                    web.get('/information', index_handler),
                    web.get('/control', index_handler),
                    web.get('/records', index_handler),
                    web.get('/logging', index_handler),
                    web.get('/bin', index_handler),
                    web.get('/ws', ws_comm_handler.receive)])
    # From the aiohttp documentation it is known to use
    # add_static only when developing things
    # normally static content should be processed by
    # webservers like (nginx or apache)
    # In the case of Single Tester it is okay to use add_static
    app.router.add_static('/', path=static_file_path, name='static')
    app['ws_comm_handler'] = ws_comm_handler


async def webservice_cleanup(app):
    ws_comm_handler = app['ws_comm_handler']
    if ws_comm_handler is not None:
        await ws_comm_handler.close_all()
        app['ws_comm_handler'] = None


def webservice_setup_app(app: web.Application, static_file_path: str):
    if not os.path.isdir(static_file_path):
        raise ValueError(f'static_file_path is not an existing directory: {static_file_path}')
    
    app['static_file_path'] = static_file_path
    app.on_startup.append(webservice_init)
    app.on_cleanup.append(webservice_cleanup)


class WebSocketConnection:
    def __init__(self, ws_connection, connection_id):
        self.ws_connection = ws_connection
        self.connection_id = connection_id

    def is_alive(self):
        return not (self.ws_connection.closed or self.ws_connection.close_code is not None)

    def close(self, code, message):
        self.ws_connection.close(code=code, message=message)

    def get_connection_id(self):
        return self.connection_id

    async def send_json(self, data):
        await self.ws_connection.send_json(data)



from ate_common.program_utils import BinTableFieldName
from dataclasses import dataclass


@dataclass()
class BinInfo:
    bin_name: str = ''
    sbin: int = -1
    hbin: int = -1
    count: int = 0
    result_flag: str = ''
    head_num: int = 0

    def update_infos(self, result_flag: str, head_num: int):
        self.count += 1
        self.result_flag = 'P' if result_flag == 0 else 'F'
        self.head_num = head_num

    def is_bin_name(self, bin_name):
        return bin_name == self.bin_name

    def to_dict(self):
        return {self.bin_name: [self.sbin, self.hbin, self.count, self.result_flag, self.head_num]}


class SiteInformation:
    def __init__(self, site_id: int, site_infos: dict):
        self.site_id = site_id
        self.site_infos = self.generate_site_infos(site_infos)

    def store_test_result(self, prr_record: dict):
        self.site_infos[str(prr_record['SOFT_BIN'])].update_infos(prr_record['PART_FLG'], int(prr_record['HEAD_NUM']))

    @staticmethod
    def generate_site_infos(site_infos: dict) -> dict:
        ret_dict = {}
        for site_info in site_infos:
            ret_dict[site_info[BinTableFieldName.SBinNum()]] =\
                BinInfo(site_info[BinTableFieldName.SBinName()],
                        int(site_info[BinTableFieldName.SBinNum()]),
                        int(site_info[BinTableFieldName.HBin()]),
                        0,
                        'F')

        return ret_dict

    def get_bin_information(self) -> dict:
        ret_dict = {}
        for _, site_info in self.site_infos.items():
            ret_dict.update(site_info.to_dict())

        return ret_dict


class BinInformationHandler:
    def __init__(self, sites_id: list):
        self.sites_info = {}
        self.sites_id = sites_id
        self.bin_names = []

    def set_sites_information(self, site_infos: dict):
        self.bin_names = [site_info[BinTableFieldName.SBinName()] for site_info in site_infos]

        for site_id in self.sites_id:
            self.sites_info[site_id] = SiteInformation(site_id, site_infos)

    def collect_test_result(self, prr_records: dict):
        for _, prr_record in prr_records.items():
            site_id = str(prr_record['SITE_NUM'])
            self.sites_info[site_id].store_test_result(prr_record)

    def get_summary_information(self, prr_records: dict) -> dict:
        self.collect_test_result(prr_records)
        summary = {}
        for site_id, site_info in self.sites_info.items():
            bin_info = site_info.get_bin_information()

            for bin_name in self.bin_names:
                summary.setdefault(bin_name, {}).update({site_id: bin_info[bin_name]})

        return summary

    def get_num_sites(self) -> int:
        return len(self.sites_info)

    def clear_site_information(self):
        self.sites_info.clear()



from ate_apps_common.mqtt_connection import MqttConnection
from ate_common.logger import LogLevel
import json
import re
from typing import List, Optional

TOPIC_CONTROLSTATUS = "Control/status"
TOPIC_CONTROLLOG = "Control/log"
SUBTOPIC_HANDLER = "Handler/status"

INTERFACE_VERSION = 1


class MasterConnectionHandler:

    """ handle commands """

    def __init__(self, host, port, sites, device_id, handler_id, status_consumer):
        mqtt_client_id = f'masterapp.{device_id}'
        self.status_consumer = status_consumer
        self.log = status_consumer.log
        self.mqtt = MqttConnection(host, port, mqtt_client_id, self.log)
        self.mqtt.init_mqtt_client_callbacks(self._on_connect_handler,
                                             self._on_disconnect_handler)
        self.sites = sites
        self.device_id = device_id
        self.connected_flag = False
        self.handler_id = handler_id

        self.mqtt.register_route("Control", lambda topic, payload: self.dispatch_control_message(topic, self.mqtt.decode_payload(payload)))
        self.mqtt.register_route("TestApp", lambda topic, payload: self.dispatch_testapp_message(topic, self.mqtt.decode_payload(payload)))
        self.mqtt.register_route("Master/cmd", lambda topic, payload: self.dispatch_handler_message(topic, self.mqtt.decode_payload(payload)))
        self.mqtt.register_route("Handler", lambda topic, payload: self.dispatch_handler_message(topic, self.mqtt.decode_payload(payload)))

    def start(self):
        self.mqtt.set_last_will(
            self._generate_base_topic_status(),
            self.mqtt.create_message(
                self._generate_status_message('crash')))
        self.mqtt.start_loop()

    async def stop(self):
        await self.mqtt.stop_loop()

    def publish_state(self, state, statedict=None):
        self.mqtt.publish(self._generate_base_topic_status(),
                          self.mqtt.create_message(self._generate_status_message(state, statedict)),
                          qos=2,
                          retain=False)

    def send_load_test_to_all_sites(self, testapp_params):
        topic = f'ate/{self.device_id}/Control/cmd'
        params = {
            'type': 'cmd',
            'command': 'loadTest',
            'testapp_params': testapp_params,
            'sites': self.sites,
        }
        self.log.log_message(LogLevel.Info(), 'Send LoadLot to sites...')
        self.mqtt.publish(topic, json.dumps(params), 2, False)

    def send_next_to_all_sites(self, job_data: Optional[dict] = None):
        topic = f'ate/{self.device_id}/TestApp/cmd'
        params = self._generate_command_message('next', job_data['sites'])

        if job_data is not None:
            params['job_data'] = job_data
            job_data.pop('sites')
        self.mqtt.publish(topic, json.dumps(params), 2, False)

    def send_terminate_to_all_sites(self):
        topic = f'ate/{self.device_id}/TestApp/cmd'
        self.mqtt.publish(topic, json.dumps(self._generate_command_message('terminate')))

    def send_test_results(self, test_results):
        topic = f'ate/{self.device_id}/Master/response'
        self.mqtt.publish(topic, json.dumps(self._generate_test_results_message(test_results)))

    def send_identification(self):
        topic = f'ate/{self.device_id}/Master/response'
        self.mqtt.publish(topic, json.dumps(self._generate_identification_message()))

    def send_host_info(self, host: str, port: int):
        topic = f'ate/{self.device_id}/Master/response'
        self.mqtt.publish(topic, json.dumps(self._generate_host_message(host, port)))

    def send_state(self, state, message):
        topic = f'ate/{self.device_id}/Master/response'
        self.mqtt.publish(topic, json.dumps(self._generate_state_message(state, message)))

    def send_reset_to_all_sites(self):
        control_topic = f'ate/{self.device_id}/Control/cmd'
        self.mqtt.publish(control_topic, json.dumps(self._generate_command_message('reset')))

        # hack: this make sure to terminate any zombie test application that do not have a parent process(control)
        testApp_topic = f'ate/{self.device_id}/TestApp/cmd'
        self.mqtt.publish(testApp_topic, json.dumps(self._generate_command_message('reset')))

    def send_set_log_level(self, level):
        topics = [f'ate/{self.device_id}/TestApp/cmd', f'ate/{self.device_id}/Control/cmd']
        default_message = self._generate_command_message('setloglevel')
        default_message.update({'level': level})
        for topic in topics:
            self.mqtt.publish(topic, json.dumps(default_message))

    def send_set_new_hbin(self, sbin: int, hbin: int):
        testApp_topic = f'ate/{self.device_id}/TestApp/cmd'
        message = self._generate_command_message('sethbin')
        message.update({'sbin': sbin})
        message.update({'hbin': hbin})

        self.mqtt.publish(testApp_topic, json.dumps(message))

    def send_handler_get_temperature_command(self):
        message = self._generate_message('temperature', {})
        self.send_handler_command(message)

    def send_handler_command(self, message):
        topic = f'ate/{self.device_id}/Handler/command'
        self.mqtt.publish(topic, json.dumps(message))

    def send_get_execution_strategy_command(self, layout):
        topic = f'ate/{self.device_id}/TestApp/cmd'
        message = self._generate_command_message('getexecutionstrategy')
        message.update({'layout': layout})
        self.mqtt.publish(topic, json.dumps(message))

    def _generate_command_message(self, command: str, sites: List[str] = None):
        return {'type': 'cmd',
                'command': command,
                'sites': self.sites if sites is None else sites,
                }

    def _generate_test_results_message(self, test_results):
        return self._generate_message('next', {'sites': test_results})

    def _generate_identification_message(self):
        return self._generate_message('identify', {'name': self.device_id})

    def _generate_host_message(self, host: str, port: int):
        return self._generate_message('get-host', {'host': host, 'port': port})

    def _generate_state_message(self, state, message):
        return self._generate_message('get-state', {'state': state, 'message': message})

    @staticmethod
    def _generate_message(type, payload):
        return {'type': type, 'payload': payload}

    def on_cmd_message(self, message):
        payload = self.decode_message(message)

        to_exec_command = self.commands.get(payload.get("value"))
        if to_exec_command is None:
            self.log.log_message(LogLevel.Warning(), 'received command not found')
            return False

        return to_exec_command()

    def _on_connect_handler(self, client, userdata, flags, conect_res):
        self.log.log_message(LogLevel.Info(), 'mqtt connected')
        self.connected_flag = True

        self.mqtt.subscribe(f"ate/{self.device_id}/Master/#")
        self.mqtt.subscribe(f"ate/{self.device_id}/TestApp/#")
        self.mqtt.subscribe(f"ate/{self.device_id}/Control/#")
        self.mqtt.subscribe(f"ate/{self.handler_id}/{SUBTOPIC_HANDLER}")
        self.mqtt.subscribe(f"ate/{self.device_id}/+/log/#")

        self.status_consumer.startup_done()
        self.send_reset_to_all_sites()

    def _on_disconnect_handler(self, client, userdata, distc_res):
        self.log.log_message(LogLevel.Info(), f'mqtt disconnected (rc: {distc_res})')
        self.connected_flag = False

    def _generate_status_message(self, state, message=''):
        message = {
            "type": "status",
            "interface_version": INTERFACE_VERSION,
            "state": state,
            "payload": {"state": state, "message": message}
        }
        return message

    def _generate_base_topic_status(self):
        return "ate/" + str(self.device_id) + "/Master/status"

    def __generate_ioctl_response_topic(self, resource_request):
        return f"ate/{self.device_id}/Master/{resource_request['periphery_type']}/response"

    def publish_ioctl_response(self, resource_request, result):
        response_topic = self.__generate_ioctl_response_topic(resource_request)
        self.mqtt.publish(response_topic, result, 2)

    def publish_ioctl_timeout(self, resource_request):
        response_topic = self.__generate_ioctl_response_topic(resource_request)
        message = {
            "type": "io-control-response",
            "ioctl_name": resource_request["ioctl_name"],
            "result": "Timeout"
        }
        self.mqtt.publish(response_topic, json.dumps(message), 2)

    def _generate_usersettings_message(self, usersettings: dict):
        message = {
            "type": "usersettings",
        }
        message.update(usersettings)
        return message

    def __extract_siteid_from_control_topic(self, topic):
        pats = [rf'ate/{self.device_id}/{TOPIC_CONTROLSTATUS}/site(.+)$',
                rf'ate/{self.device_id}/{TOPIC_CONTROLLOG}/site(.+)$']

        for pat in pats:
            m = re.match(pat, topic)
            if m:
                return m.group(1)

    def __extract_siteid_from_testapp_topic(self, topic):
        patterns = [rf'ate/{self.device_id}/TestApp/(?:status|testresult)/site(.+)$',
                    rf'ate/{self.device_id}/TestApp/(?:status|testsummary|log|execution_strategy)/site(.+)$',
                    rf'ate/{self.device_id}/TestApp/io-control/site(.+)/request$',
                    rf'ate/{self.device_id}/TestApp/binsettings/site(.+)$']
        for pattern in patterns:
            m = re.match(pattern, topic)
            if m:
                return m.group(1)

    def dispatch_control_message(self, topic, msg):
        siteid = self.__extract_siteid_from_control_topic(topic)
        if siteid is None:
            # Hacky: To limit the number of routs we subscribed to # on control, thus we get served
            # the messages in cmd as well, which don't have a siteid in the topicname. To avoid
            # confusing logmessage we return early in that case:
            if "cmd" in topic:
                return
            self.log.log_message(LogLevel.Warning(), f'unexpected message on control topic {topic}: extracting siteid failed')
            return

        if "status" in topic:
            self.status_consumer.on_control_status_changed(siteid, msg)
        elif "log" in topic:
            self.status_consumer.on_log_message(siteid, msg)
        else:
            assert False

    def dispatch_testapp_message(self, topic, msg):
        siteid = self.__extract_siteid_from_testapp_topic(topic)
        if siteid is None:
            # Hacky: To limit the number of routs we subscribed to # on control, thus we get served
            # the messages in cmd as well, which don't have a siteid in the topicname. To avoid
            # confusing logmessage we return early in that case:
            if "cmd" in topic:
                return
            self.log.log_message(LogLevel.Warning(), f'unexpected message on testapp topic {topic}: extracting siteid failed')
            return

        if "testresult" in topic:
            self.status_consumer.on_testapp_testresult_changed(siteid, msg)
        elif "testsummary" in topic:
            self.status_consumer.on_testapp_testsummary_changed(msg)
        elif "io-control" in topic:
            assert 'type' in msg
            assert msg['type'] == 'io-control-request'
            assert 'periphery_type' in msg
            assert 'ioctl_name' in msg
            assert 'parameters' in msg
            self.status_consumer.on_testapp_resource_changed(siteid, msg)
        elif "status" in topic:
            self.status_consumer.on_testapp_status_changed(siteid, msg)
        elif "log" in topic:
            self.status_consumer.on_log_message(siteid, msg)
        elif "execution_strategy":
            self.status_consumer.on_execution_strategy_message(siteid, msg['payload'])
        else:
            assert False

    def dispatch_handler_message(self, topic, msg):
        if "status" in topic:
            self.status_consumer.on_handler_status_changed(msg["payload"])
        elif "cmd" in topic:
            self.status_consumer.on_handler_command_message(msg)
        elif "response" in topic:
            self.status_consumer.on_handler_response_message(msg)
        else:
            assert False



from ate_master_app.utils.bin_information import BinInformationHandler
from ate_common.program_utils import BinTableFieldName
from ate_master_app.utils.part_information import PartInformationHandler
from ate_master_app.utils.yield_information import YieldInformationHandler
from ate_master_app.utils.table_information import BinTableInformationHandler

REQUIRED_FIELDS = ['HEAD_NUM', 'SITE_NUM', 'HARD_BIN', 'SOFT_BIN', 'PART_ID', 'PART_RETEST', 'PART_FLG']


class ResultInformationHandler:
    def __init__(self, sites_id: list):
        self._yield_info_handler = YieldInformationHandler()
        self._part_info_handler = PartInformationHandler()
        self._site_info_handler = BinInformationHandler(sites_id)
        self._bin_table_info_handler = BinTableInformationHandler(sites_id)
        self.prr_rec_information = {}

    def handle_result(self, prr_record: dict):
        part_id = self._get_part_id(prr_record)
        site_num = str(prr_record['SITE_NUM'])
        soft_bin = str(prr_record['SOFT_BIN'])
        message_handle_result = (True, '')

        # add 'PART_RETEST' field to count part retest commands
        if not self.prr_rec_information.get(part_id):
            prr_record['PART_RETEST'] = 0
            self.prr_rec_information[part_id] = self._extract_needed_fields(prr_record)
            message_handle_result = self._yield_info_handler.accumulate_site_bin_info(site_num, soft_bin)
            self._bin_table_info_handler.accumulate_bin_table_info(site_num, soft_bin)
        else:
            old_prr_rec = self.prr_rec_information[part_id]
            part_count = old_prr_rec['PART_RETEST']
            prr_record['PART_RETEST'] = part_count + 1
            self.prr_rec_information.update({part_id: prr_record})
            message_handle_result = self._yield_info_handler.reaccumulate_bin_info(self.prr_rec_information)
            self._bin_table_info_handler.reaccumulate_bin_table_info(self.prr_rec_information)

        return message_handle_result

    @staticmethod
    def _extract_needed_fields(prr: dict):
        return {field_name: prr[field_name] for field_name in REQUIRED_FIELDS}

    @staticmethod
    def _get_part_id(prr_record: dict):
        part_id = ''
        if prr_record['PART_ID']:
            part_id = prr_record['PART_ID']
        else:
            default_value = -32768
            if (prr_record['X_COORD'], prr_record['Y_COORD']) == (default_value, default_value):
                assert False
            else:
                part_id = f"{prr_record['X_COORD']}x{prr_record['Y_COORD']}"

        return part_id

    def set_bin_settings(self, bin_table: dict):
        bin_settings = self.get_bin_settings(bin_table)

        self._site_info_handler.set_sites_information(bin_table)
        self._yield_info_handler.set_bin_settings(bin_settings)
        self._part_info_handler.set_bin_settings(bin_settings)
        self._bin_table_info_handler.set_bin_table(bin_table)

    def update_hbin_number(self, sbin: str, hbin: str):
        self._bin_table_info_handler._update_hbin_num(sbin, hbin)

    @staticmethod
    def get_bin_settings(bin_table: list) -> dict:
        settings = {}

        for bin_info in bin_table:
            settings.setdefault(bin_info[BinTableFieldName.SBinGroup()], {'hbins': [], 'sbins': []})['hbins'].append(bin_info[BinTableFieldName.HBin()])
            settings.setdefault(bin_info[BinTableFieldName.SBinGroup()], {'hbins': [], 'sbins': []})['sbins'].append(bin_info[BinTableFieldName.SBinNum()])

        return settings

    def clear_all(self):
        self._yield_info_handler.clear_yield_information()
        self._part_info_handler.clear_part_information()
        self._site_info_handler.clear_site_information()
        self._bin_table_info_handler.clear_bin_table()
        self.prr_rec_information.clear()

    def get_site_result_response(self, prr_record: dict) -> dict:
        return self._yield_info_handler.get_site_result(prr_record)

    def get_yield_messages(self) -> dict:
        return self._yield_info_handler.get_yield_messages()

    def get_site_yield_info_message(self, site: str) -> dict:
        return self._yield_info_handler.get_site_yield_info_message(site)

    def get_hbin_soft_bin_report(self) -> dict:
        return self._site_info_handler.get_summary_information(self.prr_rec_information)

    def get_part_count_infos(self) -> dict:
        return self._part_info_handler.get_part_count_infos(self.prr_rec_information)

    def get_bin_table(self) -> list:
        return self._bin_table_info_handler.get_bin_table_infos()



import json


class UserSettings:

    @staticmethod
    def get_defaults() -> dict:
        # TODO: do we need this
        # 'master.disabled_site_ids': []
        return {
            'stop_on_fail': {'active': False, 'value': -1},
            'single_step': {'active': False, 'value': -1},
            'stop_on_test': {'active': False, 'value': -1},
            'trigger_on_test': {'active': False, 'value': -1},
            'trigger_on_fail': {'active': False, 'value': -1},
            'trigger_site_specific': {'active': False, 'value': -1},
        }

    @staticmethod
    def read_settings_file(path: str) -> dict:
        with open(path) as f:
            json_value = json.load(f)
            if not isinstance(json_value, dict):
                raise ValueError(f'invalid settings file "{path}": content must be a json object, but was "{type(json_value)}"')
            return json_value

    @classmethod
    def write_settings_file(cls, path: str, settings: dict):
        with open(path, 'w') as f:
            json.dump(settings, f, indent=4, sort_keys=True)

    @classmethod
    def load_from_file(cls, path: str) -> dict:
        settings = cls.get_defaults()
        settings_from_file = cls.read_settings_file(path)
        settings.update(settings_from_file)
        return settings

    @classmethod
    def save_to_file(cls, path: str, settings: dict, add_defaults: bool = True):
        settings_to_write = cls.get_defaults() if add_defaults else {}
        settings_to_write.update(settings)
        cls.write_settings_file(path, settings_to_write)



class BinTableInformationHandler:
    def __init__(self, sites: list):
        self._bin_table = {}
        self._sites = sites

    def set_bin_table(self, bin_table: dict):
        for bin_info in bin_table:
            self._bin_table[bin_info['SBIN']] = self._generate_table_row(bin_info['SBINNAME'], int(bin_info['SBIN']), int(bin_info['HBIN']), bin_info['GROUP'])

    def _generate_table_row(self, bin_name: str, sbin: int, hbin: int, typ: str):
        site_counts = []
        for site in self._sites:
            site_counts.append({'siteId': site, 'count': 0})

        return {'name': bin_name, 'sBin': sbin, 'hBin': hbin, 'type': typ, 'siteCounts': site_counts}

    def _update_hbin_num(self, sbin: str, hbin: str):
        self._bin_table[sbin]['hBin'] = int(hbin)

    def clear_bin_table(self):
        self._bin_table.clear()

    def get_bin_table_infos(self) -> list:
        return [table_row for _, table_row in self._bin_table.items()]

    def accumulate_bin_table_info(self, site_id: str, sbin: int):
        return self._update_site_count(site_id, sbin)

    def _update_site_count(self, site_id: str, sbin: int):
        for site_count in self._bin_table[str(sbin)]['siteCounts']:
            if site_count['siteId'] != site_id:
                continue

            site_count['count'] += 1

    def reaccumulate_bin_table_info(self, prrs: list):
        self._reset_site_count()
        for _, prr in prrs.items():
            site = str(prr['SITE_NUM'])
            sbin = str(prr['SOFT_BIN'])
            self._update_site_count(site, int(sbin))

    def _reset_site_count(self):
        for _, bin_info in self._bin_table.items():
            for site_count in bin_info['siteCounts']:
                site_count['count'] = 0



from typing import List
from abc import ABC, abstractmethod


def get_execution_strategy(strategy_type: str, tester: object):
    if 'simple' == strategy_type:
        return SimpleExecutionStrategy(tester)
    elif 'default' == strategy_type:
        return DefaultExecutionStrategy(tester)

    assert False, f"strategy with type: {strategy_type} is not supported"


class IExecutionStrategy(ABC):
    @abstractmethod
    def handle_release(self, sites: List[str]) -> List[str]:
        pass

    @abstractmethod
    def reset_stages(self):
        pass

    @abstractmethod
    def set_configuration(self, strategy_config: list):
        pass

    @abstractmethod
    async def get_site_states(self, timeout: int) -> List[int]:
        pass


class SimpleExecutionStrategy(IExecutionStrategy):
    def __init__(self, tester: object) -> None:
        self.tester = tester

    @staticmethod
    def handle_release(sites: List[str]) -> List[str]:
        return sites

    @staticmethod
    def reset_stages():
        pass

    @staticmethod
    def set_configuration(strategy_config: dict):
        pass

    async def get_site_states(self, timeout: int) -> List[int]:
        return await self.tester.get_site_states(timeout)


class DefaultExecutionStrategy:
    def __init__(self, tester: object):
        self._execution_strategy = {}
        self._stage_exectution_strategy = []
        self.tester = tester
        self._test_num = -1

    def handle_release(self, sites: List[str]) -> List[str]:
        if not len(sites):
            raise Exception('DefaultExecutionStrategy cannot release empty site list')

        sites_to_test = self.get_strategy_for_sites(sites)
        self.tester.release_test_execution(sites_to_test)
        return sites_to_test

    def get_strategy_for_sites(self, sites: List[str]) -> List[str]:
        if not self._stage_exectution_strategy:
            self._test_num += 1

            self._stage_exectution_strategy = self._get_testing_sites(sites)

        if not self._stage_exectution_strategy:
            raise Exception('DefaultExecutionStrategy cannot release empty site list')

        return self._stage_exectution_strategy.pop(0)

    def _get_testing_sites(self, sites: List[str]) -> List[str]:
        if self._test_num == -1:
            return []

        if not self._execution_strategy:
            raise Exception('execution strategy is not configured yet, no configuration is received')

        import copy
        stages = copy.deepcopy(self._execution_strategy[self.test_num])

        # all sites not contained in the list of sites that should test,
        # must be removed from the execution strategy
        for stage in stages:
            for index, site in enumerate(stage):
                if site in sites:
                    continue

                stage.pop(index)
                break

        # return all non empty list
        return [stage for stage in stages if len(stage)]

    async def get_site_states(self, timeout: int) -> List[int]:
        return await self.tester.get_site_states(timeout)

    def _is_valid_test_num(self, test_num: int) -> bool:
        return test_num in range(0, len(self._execution_strategy))

    def set_configuration(self, strategy_config: dict):
        self._execution_strategy = strategy_config

    @property
    def test_num(self) -> int:
        return self._test_num

    @property
    def sites_to_release(self) -> List[List[str]]:
        return self._stage_exectution_strategy

    def reset_stages(self):
        if len(self._stage_exectution_strategy):
            raise Exception(f'cannot reset execution_strategy, the following stages are not handled yet: {self._stage_exectution_strategy}')

        self._test_num = -1



class StaticDataSource:
    def __init__(self):
        pass

    def retrieve_data(self) -> dict:
        return {}

    def verify_data(self, data) -> bool:
        return data is not None

    def get_test_information(self, data) -> dict:
        return data

    @staticmethod
    def get_bin_table(data: list) -> list:
        return data

    @staticmethod
    def get_binning_tuple(bin_table: list) -> dict:
        return bin_table

    @staticmethod
    def get_bin_settings(bin_table: list) -> dict:
        return bin_table



from ate_common.program_utils import ALARM_BIN_MAX, ALARM_BIN_MIN, GOOD_BIN_MAX, GOOD_BIN_MIN


class PartInformation:
    def __init__(self, head_num, site_num):
        self.part_count = 0
        self.retest_count = 0
        self.abort_count = 0
        self.good_count = 0
        self.functional_count = 0
        self.head_num = head_num
        self.site_num = site_num

    def update_part_information(self, prr_record):
        self.part_count += 1
        self.retest_count += prr_record['PART_RETEST']

        soft_bin = prr_record['SOFT_BIN']
        if self._does_abnormally_end(soft_bin):
            self.abort_count += 1

        if self._is_good_bin(soft_bin):
            self.good_count += 1
            self.functional_count += 1

    @staticmethod
    def _does_abnormally_end(soft_bin: int) -> bool:
        return soft_bin in range(ALARM_BIN_MIN, ALARM_BIN_MAX + 1)

    @staticmethod
    def _is_good_bin(soft_bin: int) -> bool:
        return soft_bin in range(GOOD_BIN_MIN, GOOD_BIN_MAX + 1)

    def get_part_count_info(self) -> dict:
        return {'head_num': int(self.head_num), 'site_num': int(self.site_num), 'part_count': self.part_count,
                'retest_count': self.retest_count, 'abort_count': self.abort_count, 'good_count': self.good_count,
                'functional_count': self.functional_count}


class PartInformationHandler:
    def __init__(self):
        self.part_information = {}

    def set_bin_settings(self, bin_settings: dict):
        self._bin_settings = bin_settings

    def update_part_count_records(self, prr_records: dict):
        for _, prr_record in prr_records.items():
            siteid = str(prr_record['SITE_NUM'])

            if not self.part_information.get(siteid):
                head_num = str(prr_record['HEAD_NUM'])
                self.part_information[siteid] = PartInformation(head_num, siteid)

            self.part_information[siteid].update_part_information(prr_record)

    def get_part_count_infos(self, prr_records: dict) -> list:
        self.update_part_count_records(prr_records)
        part_infos = []
        for _, part_info in self.part_information.items():
            part_infos.append(part_info.get_part_count_info())

        part_infos.append(self._generate_summary_part_count_info(part_infos))
        return part_infos

    @staticmethod
    def _generate_summary_part_count_info(part_infos: list) -> dict:
        summary_part_count_info = {'head_num': 255, 'site_num': 255, 'part_count': 0,
                                   'retest_count': 0, 'abort_count': 0, 'good_count': 0,
                                   'functional_count': 0}

        for part_info in part_infos:
            summary_part_count_info['part_count'] += part_info['part_count']
            summary_part_count_info['retest_count'] += part_info['retest_count']
            summary_part_count_info['abort_count'] += part_info['abort_count']
            summary_part_count_info['good_count'] += part_info['good_count']
            summary_part_count_info['functional_count'] += part_info['functional_count']

        return summary_part_count_info

    def clear_part_information(self):
        self.part_information.clear()



from transitions.extensions import HierarchicalMachine as Machine


class TestsExecutionMachine(Machine):
    states = ['idle', 'busy', 'terminate', 'waiting_for_release', 'waiting_for_resource', 'error']

    def __init__(self, model=None):
        super().__init__(model=model, states=self.states, initial='idle')

        # startup:
        self.add_transition(trigger='on_release',           source='idle',                  dest='waiting_for_release')      # noqa: E241
        self.add_transition(trigger='on_release',           source='waiting_for_release',   dest='busy')      # noqa: E241

        # request:
        self.add_transition(trigger='on_test_request',      source='busy',                  dest='waiting_for_release')   # noqa: E241
        self.add_transition(trigger='on_reqesut_release',   source='waiting_for_release',   dest='busy')         # noqa: E241

        # resource:
        self.add_transition(trigger='on_resource_request',  source='busy',                  dest='waiting_for_resource')          # noqa: E241
        self.add_transition(trigger='on_resource_ready',    source='waiting_for_resource',  dest='busy')    # noqa: E241

        # on error
        self.add_transition(trigger='on_error',             source='*',                     dest='error')          # noqa: E241



import xmltodict


class XmlParameterParser:

    def __init__(self):
        self.xml_data = None
        self.root = None
        self.param_dict = {"MAIN": None,
                           "CLUSTER": None,
                           "TESTPROGRAM": None,
                           "HANDLERSECTION": None,
                           "PAT": None,
                           "STATION": None,
                           "DIEELEMENT_COUNT": None,
                           "DIEELEMENT": None}

    def parse(self, path: str) -> dict:
        with open(path) as x:
            param = xmltodict.parse(x.read())
            self.root = next(iter(param))

        self.param_dict.update(param.get(self.root))
        return self.param_dict



""" master App """

# External imports
from transitions.core import MachineError
from transitions.extensions import HierarchicalMachine as Machine
from queue import Empty, Full, Queue
import asyncio
import mimetypes
import sys

from typing import Callable, List

from ate_common.logger import Logger, LogLevel
from ate_apps_common.sequence_container import SequenceContainer
from ate_master_app.master_connection_handler import MasterConnectionHandler
from ate_master_app.parameter_parser import parser_factory
from ate_master_app.user_settings import UserSettings
from ate_master_app.utils.command_executor import (GetLogFileCommand, GetLogsCommand, GetLotData, GetTestResultsCommand, GetUserSettings,
                                                                  GetYields, GetBinTable)
from ate_master_app.task_handler.task_handler import TaskHandler

from ate_apps_common.stdf_aggregator import StdfTestResultAggregator
from ate_master_app.resulthandling.result_collection_handler import ResultsCollector
from ate_master_app.utils.result_Information_handler import ResultInformationHandler

from ate_master_app.peripheral_controller import PeripheralController

from ate_master_app.statemachines.MultiSite import (MultiSiteTestingModel, MultiSiteTestingMachine)
from ate_master_app.execution_strategy.execution_strategy import IExecutionStrategy, get_execution_strategy
from os import getcwd
from pathlib import Path


INTERFACE_VERSION = 1
MAX_NUM_OF_TEST_PROGRAM_RESULTS = 1000


def assert_valid_system_mimetypes_config():
    """
    Perform sanity check for system/enfironment configuration and return
    result as boolean.

    Background info:

    aiohttp uses mimetypes.guess_type() to guess the content-type to be
    used in the http response headers when serving static files.

    If we serve javascript modules with "text/plain" instead
    of "application/javascript" the browser will not execute the file as
    javascript module and the angular frontend does not load.

    On windows mimetypes.init (called automatically by guess_type if not
    called before) will always read content types from registry in Python
    3.8.1 (e.g. "HKLM\\Software\\Classes\\.js\\Content Type"). The values
    stored there may not be standard because they have been changed on
    certain systems (reasons unknown).

    Apparently it was possible to avoid this in earlier python
    version by explicitly passing empty list to files, e.g.
    mimetypes.init(files=[]). But this does not work anymore in 3.8.1,
    where types from registry will always be loaded.
    """
    js_mime_type = mimetypes.guess_type('file.js')[0]
    if js_mime_type != 'application/javascript':
        print('FATAL ERROR: Invalid system configuration for .js type: '
              + 'expected "application/javascript" but got '
              + f'"{js_mime_type}".'
              + ' Please fix your systems mimetypes configuration.')
        sys.exit(1)


CONTROL_STATE_UNKNOWN = "unknown"
CONTROL_STATE_LOADING = "loading"
CONTROL_STATE_BUSY = "busy"
CONTROL_STATE_IDLE = "idle"
CONTROL_STATE_CRASH = "crash"

TEST_STATE_IDLE = "idle"
TEST_STATE_TESTING = "testing"
TEST_STATE_CRASH = "crash"
TEST_STATE_TERMINATED = "terminated"

STARTUP_TIMEOUT = 300
LOAD_TIMEOUT = 180
UNLOAD_TIMEOUT = 60
TEST_TIMEOUT = 30
RESET_TIMEOUT = 20
RESPONSE_TIMEOUT = 10


class MasterApplication(MultiSiteTestingModel):

    states = ['startup',
              'connecting',
              'initialized',
              'loading',
              'ready',
              {'name': 'testing', 'children': MultiSiteTestingMachine()},  # , 'remap': {'completed': 'ready'}
              'finished',
              'unloading',
              'error',
              'softerror',
              'waitingforbintable']

    # multiple space code style "error" will be ignored for a better presentation of the possible state machine transitions
    transitions = [
        {'source': 'startup',            'dest': 'connecting',  'trigger': "startup_done",                           'after': "on_startup_done"},               # noqa: E241
        {'source': 'connecting',         'dest': 'initialized', 'trigger': 'all_sites_detected',                     'after': "on_allsitesdetected"},           # noqa: E241
        {'source': 'connecting',         'dest': 'error',       'trigger': 'bad_interface_version'},                                                            # noqa: E241

        {'source': 'initialized',        'dest': 'loading',     'trigger': 'load_command',                           'after': 'on_loadcommand_issued'},         # noqa: E241
        {'source': 'loading',            'dest': 'ready',       'trigger': 'all_siteloads_complete',                 'after': 'on_allsiteloads_complete'},      # noqa: E241
        # {'source': 'loading',            'dest': 'waiting_for_execution_strategy',    'trigger': 'all_sitesendexecution_complete',         'after': 'on_allsitesendexecution_complete'},      # noqa: E241
        # {'source': 'waiting_for_execution_strategy',     'dest': 'ready',             'trigger': 'all_siteloads_complete',                 'after': 'on_allsiteloads_complete'},      # noqa: E241
        {'source': 'loading',            'dest': 'initialized', 'trigger': 'load_error',                             'after': 'on_load_error'},                 # noqa: E241

        # TODO: properly limit source states to valid states where usersettings are allowed to be modified
        #       ATE-104 says it should not be possible while testing in case of stop-on-fail,
        #       but this constraint may not be required here and could be done in UI)
        {'source': ['initialized', 'ready'], 'dest': '=', 'trigger': 'usersettings_command',                       'after': 'on_usersettings_command_issued'},  # noqa: E241

        {'source': 'ready',              'dest': 'testing',                'trigger': 'next',                      'after': 'on_next_command_issued'},          # noqa: E241
        {'source': 'ready',              'dest': 'unloading',              'trigger': 'unload',                    'after': 'on_unload_command_issued'},        # noqa: E241
        {'source': 'testing_completed',  'dest': 'ready',                  'trigger': 'all_sitetests_complete',    'after': "on_allsitetestscomplete"},         # noqa: E241
        {'source': 'unloading',          'dest': 'initialized',            'trigger': 'all_siteunloads_complete',  'after': "on_allsiteunloadscomplete"},       # noqa: E241

        {'source': 'ready',              'dest': '=',                      'trigger': 'getresults',                'after': 'on_getresults_command'},           # noqa: E241
        {'source': '*',                  'dest': '=',                      'trigger': 'getlogs',                   'after': 'on_getlogs_command'},              # noqa: E241
        {'source': '*',                  'dest': '=',                      'trigger': 'getlogfile',                'after': 'on_getlogfile_command'},           # noqa: E241
        {'source': '*',                  'dest': '=',                      'trigger': 'setloglevel',               'after': 'on_setloglevel_command'},          # noqa: E241

        {'source': '*',                  'dest': 'softerror',              'trigger': 'testapp_disconnected',      'after': 'on_disconnect_error'},             # noqa: E241
        {'source': '*',                  'dest': 'softerror',              'trigger': 'timeout',                   'after': 'on_timeout'},                      # noqa: E241
        {'source': '*',                  'dest': 'softerror',              'trigger': 'on_error',                  'after': 'on_error_occurred'},               # noqa: E241
        {'source': 'softerror',          'dest': 'connecting',             'trigger': 'reset',                     'after': 'on_reset_received'},               # noqa: E241
        {'source': 'testing',            'dest': 'testing',                'trigger': 'all_site_request_testing',  'after': "on_all_sites_request_testing"},       # noqa: E241

        {'source': '*',                  'dest': 'ready',                  'trigger': 'all_sitetests_complete',    'after': "on_allsitetestscomplete"},         # noqa: E241
    ]

    """ MasterApplication """

    def __init__(self, configuration):
        self.sites = configuration['sites']
        super().__init__(self.sites)
        self.fsm = Machine(model=self,
                           states=MasterApplication.states,
                           transitions=MasterApplication.transitions,
                           initial="startup",
                           after_state_change='publish_state')
        self.configuration = configuration
        self.log = Logger('master')
        self.loglevel = LogLevel.Warning() if configuration.get('loglevel') is None else configuration['loglevel']
        self.log.set_logger_level(self.loglevel)
        self.apply_configuration(configuration)
        self.init()
        self.connectionHandler = MasterConnectionHandler(self.broker_host, self.broker_port, self.configuredSites, self.device_id, self.handler_id, self)
        self.peripheral_controller = PeripheralController(self.connectionHandler.mqtt, self.device_id)

        self.received_site_test_results = []
        self.received_sites_test_results = ResultsCollector(MAX_NUM_OF_TEST_PROGRAM_RESULTS)

        self.loaded_lot_number = ""
        self.error_message = ''

        self.last_published_state = ''
        self.summary_counter = 0
        self.tsr_messages = []

        self.command_queue = Queue(maxsize=50)
        self._result_info_handler = ResultInformationHandler(self.sites)

        self.test_results = []

        self.dummy_partid = 1
        self._first_part_tested = False
        self._stdf_aggregator = None
        self.testing_sites = []
        self.test_num = 0
        tester = self.get_tester(configuration['tester_type'])
        strategy_type = self.get_strategy_type(tester)
        self.testing_strategy: IExecutionStrategy = self.get_execution_strategy(strategy_type, tester)

        self.testing_event = asyncio.Event()
        self.layout = None

    @staticmethod
    def get_strategy_type(tester):
        return tester.get_strategy_type()

    @staticmethod
    def get_execution_strategy(strategy_type, tester):
        return get_execution_strategy(strategy_type, tester)

    def get_tester(self, tester_type: str):
        from ate_semiateplugins.pluginmanager import get_plugin_manager
        plugin_manager = get_plugin_manager()
        try:
            master_tester = plugin_manager.hook.get_tester_master(tester_name=tester_type)[0]
        except Exception as e:
            raise Exception(f'tester object for {tester_type} could not be load, reason: {e}')

        return master_tester

    def init(self):
        self.siteStates = {site_id: CONTROL_STATE_UNKNOWN for site_id in self.configuredSites}
        self.pendingTransitionsControl = SequenceContainer([CONTROL_STATE_IDLE], self.configuredSites, lambda: self.all_sites_detected(),
                                                           lambda site, state: self.on_unexpected_control_state(site, state))
        self.pendingTransitionsTest = SequenceContainer([TEST_STATE_IDLE], self.configuredSites, lambda: self.all_siteloads_complete(),
                                                        lambda site, state: self.on_unexpected_testapp_state(site, state))
        self.init_user_settings()

        self.timeoutHandle = None
        self.arm_timeout(STARTUP_TIMEOUT, lambda: self.timeout("Not all sites connected."))

    def apply_configuration(self, configuration: dict):
        try:
            self.configuredSites = configuration['sites']
            # Sanity check for bad configurations:
            if len(self.configuredSites) == 0:
                self.log.log_message(LogLevel.Error(), 'Master got no sites assigned')
                sys.exit()

            self.device_id = configuration['device_id']
            self.broker_host = configuration['broker_host']
            self.broker_port = configuration['broker_port']
            self.enableTimeouts = configuration['enable_timeouts']
            self.handler_id = configuration['Handler']
            self.env = configuration['environment']
        except KeyError as e:
            self.log.log_message(LogLevel.Error(), f'Master got invalid configuration: {e}')
            sys.exit()

    @property
    def user_settings_filepath(self):
        return self.configuration.get("user_settings_filepath")

    @property
    def persistent_user_settings_enabled(self):
        return self.user_settings_filepath is not None

    def init_user_settings(self):
        self.user_settings = self._load_usersettings()

    def _load_usersettings(self):
        if self.persistent_user_settings_enabled:
            try:
                user_settings = UserSettings.load_from_file(self.user_settings_filepath)
            except FileNotFoundError:
                user_settings = UserSettings.get_defaults()

            # always update file with hardcoded defaults (and create it if it does not exist)
            UserSettings.save_to_file(self.user_settings_filepath, user_settings, add_defaults=True)
        else:
            user_settings = UserSettings.get_defaults()

        return user_settings

    def modify_user_settings(self, settings):
        self.user_settings.update(self._extract_settings(settings))
        if self.persistent_user_settings_enabled:
            UserSettings.save_to_file(self.user_settings_filepath, self.user_settings, add_defaults=True)

        self.command_queue.put_nowait(GetUserSettings(lambda: self._generate_usersettings_message(self.user_settings)))

    def _store_user_settings(self, settings):
        UserSettings.save_to_file(self.user_settings_filepath, settings, add_defaults=True)
        self.user_settings = settings
        self.command_queue.put_nowait(GetUserSettings(lambda: self._generate_usersettings_message(self.user_settings)))

    def on_usersettings_command_issued(self, param_data: dict):
        settings = param_data['payload']
        self.modify_user_settings(settings)

    @staticmethod
    def _extract_settings(settings):
        modified_settings = UserSettings.get_defaults()
        for setting in settings:
            field = {setting['name']: {'active': setting['active'], 'value': int(setting['value']) if setting.get('value') else -1}}
            modified_settings.update(field)

        return modified_settings

    @property
    def external_state(self):
        return 'testing' if self.is_testing(allow_substates=True) else self.state

    def disarm_timeout(self):
        if self.enableTimeouts:
            if self.timeoutHandle is not None:
                self.timeoutHandle.cancel()
                self.timeoutHandle = None

    def arm_timeout(self, timeout_in_seconds: float, callback: Callable):
        if self.enableTimeouts:
            self.disarm_timeout()
            self.timeoutHandle = asyncio.get_event_loop().call_later(timeout_in_seconds, callback)

    def repost_state_if_connecting(self):
        # TODO: no reason to keep this ??
        if self.state == "connecting":
            self.publish_state()
            asyncio.get_event_loop().call_later(1, lambda: self.repost_state_if_connecting())

    def on_startup_done(self):
        self.repost_state_if_connecting()

    def on_timeout(self, message):
        self.error_message = message
        self.log.log_message(LogLevel.Error(), message)

    def on_disconnect_error(self, site_id, data):
        self.log.log_message(LogLevel.Error(), f'Master entered state error due to disconnect of site {site_id}')

    def on_unexpected_control_state(self, site_id, state):
        self.log.log_message(LogLevel.Warning(), f'Site {site_id} reported state {state}. This state is ignored during startup.')
        self.error_message = f'Site {site_id} reported state {state}'

        if self.state == 'connecting' and state == 'busy':
            self.on_error(f"master can't handle control {site_id}'s state 'busy' during statup")

    def on_unexpected_testapp_state(self, site_id, state):
        self.log.log_message(LogLevel.Warning(), f'TestApp for site {site_id} reported state {state}. This state is ignored during startup.')
        self.error_message = f'TestApp for site {site_id} reported state {state}'

        if self.state == 'connecting' and state == 'idle':
            self.on_error(f"master can't handle testApp {site_id}'s state 'idle' during statup")

    def on_error_occurred(self, message):
        self.log.log_message(LogLevel.Error(), f'Master entered state error, reason: {message}')
        self.error_message = message
        self.testing_event.clear()

        self.testing_strategy.reset_stages()

    def on_load_error(self):
        self.log.log_message(LogLevel.Warning(), self.error_message)

    def on_allsitesdetected(self):
        # Trap any controls that misbehave and move out of the idle state.
        # In this case we want to move to error as well
        self.pendingTransitionsControl = SequenceContainer([CONTROL_STATE_IDLE], self.configuredSites, lambda: None,
                                                           lambda site, state: self.on_error(f"Bad statetransition of control {site} during sync to {state}"))
        self.error_message = ''
        self.disarm_timeout()

    def publish_state(self, site_id=None, param_data=None):
        if self.last_published_state == self.external_state:
            return

        self.last_published_state = self.external_state
        self.log.log_message(LogLevel.Info(), f'Master state is {self.state}')
        self.connectionHandler.publish_state(self.external_state)

    def on_loadcommand_issued(self, param_data: dict):
        self._result_info_handler.clear_all()

        jobname = param_data['lot_number']
        self.loaded_lot_number = str(jobname)

        jobformat = self.configuration.get('jobformat')
        parser = parser_factory.CreateParser(jobformat)
        source = parser_factory.CreateDataSource(jobname,
                                                 self.configuration,
                                                 parser,
                                                 self.log)

        test_program_data = self.set_test_data(source)
        if not test_program_data:
            self.load_error()
            return

        self._stdf_aggregator = StdfTestResultAggregator(self.device_id + ".Master", self.loaded_lot_number, self.loaded_lot_number, self.sites)
        self._stdf_aggregator.set_test_program_data(test_program_data)

        self.arm_timeout(LOAD_TIMEOUT, lambda: self.timeout("not all sites loaded the testprogram"))
        self.pendingTransitionsControl = SequenceContainer([CONTROL_STATE_LOADING, CONTROL_STATE_BUSY], self.configuredSites, lambda: None,
                                                           lambda site, state: self.on_error(f"Bad statetransition of control {site} during load to {state}"))
        self.pendingTransitionsTest = SequenceContainer([TEST_STATE_IDLE], self.configuredSites, lambda: self.all_siteloads_complete(),
                                                        lambda site, state: self.on_error(f"Bad statetransition of testapp {site} during load to {state}"))
        self.error_message = ''

        self.connectionHandler.send_load_test_to_all_sites(self.get_test_parameters(test_program_data))
        self._store_user_settings(UserSettings.get_defaults())

        self.command_queue.put_nowait(GetBinTable(lambda: self._generate_bin_table_message()))

    def set_test_data(self, source: object):
        test_program_data = {}
        if self.configuration.get('skip_jobdata_verification', False):
            return None
        else:
            param_data = source.retrieve_data()
            if param_data is None:
                self.error_message = "Failed to execute load command"
                return None

            if not source.verify_data(param_data):
                self.error_message = "Malformed jobfile"
                return None

            test_program_data = source.get_test_information(param_data)
            bin_table = source.get_bin_table(test_program_data)
            self._result_info_handler.set_bin_settings(bin_table)

            # used to reduce the size of data to be transferred to the TP
            test_program_data['BINTABLE'] = source.get_binning_tuple(bin_table)

            self.log.log_message(LogLevel.Debug(), f'testprogram information: {test_program_data}')

        return test_program_data

    @staticmethod
    def get_test_parameters(data):
        # TODO: workaround until we specify the connection to the server or even mount the project locally
        from pathlib import Path
        import os
        return {
            'testapp_script_path': os.path.join(os.path.basename(os.fspath(Path(data['PROGRAM_DIR'])))),
            'testapp_script_args': ['--verbose'],
            'cwd': os.path.dirname(data['PROGRAM_DIR']),
            'bin_table': data['BINTABLE']                                                           # optional/unused for now
        }

    def on_allsiteloads_complete(self, paramData=None):
        self.error_message = ''
        self.disarm_timeout()

        self._send_set_log_level()
        self._send_get_execution_strategy()

    def _send_get_execution_strategy(self):
        # if handler is not connected, the tester layout should be stored
        # in the master configuration file
        if self.layout is None:
            self._set_layout(self.configuration['site_layout'])

        self.connectionHandler.send_get_execution_strategy_command(self.layout)

    def _send_set_log_level(self):
        try:
            {
                LogLevel.Debug(): 'Debug',
                LogLevel.Info(): 'Info',
                LogLevel.Warning(): 'Warning',
                LogLevel.Error(): 'Error',
            }[self.loglevel]
        except KeyError:
            self.log.log_message(LogLevel.Warning(), f'loglevel is not support: {self.loglevel}')

        self.connectionHandler.send_set_log_level(self.loglevel)

    def set_execution_strategy_configuration(self, strategy_config: List[List[List[str]]]):
        available_sites = []
        # use test_num 0 as a reference
        # make sure that all sites are contained in the list of stages
        for config in strategy_config[0]:
            available_sites.extend(config)

        if set(available_sites) != set(self.sites):
            self.on_error(f"configured sites {self.sites} are not all contained in the strategy configuration {available_sites}")

        # assuming the testprogram knew the layout of the configured sites
        self.testing_strategy.set_configuration(strategy_config)

    def on_next_command_issued(self, param_data: dict):
        if not self._first_part_tested:
            self._stdf_aggregator.write_header_records()
            self._first_part_tested = True
            self.command_queue.put_nowait(GetLotData(lambda: self._generate_lot_data_message()))

        self.received_site_test_results = []
        self.arm_timeout(TEST_TIMEOUT, lambda: self.timeout("not all sites completed the active test"))
        self.pendingTransitionsTest = SequenceContainer([TEST_STATE_TESTING, TEST_STATE_IDLE], self.configuredSites, lambda: None,
                                                        lambda site, state: self.on_error(f"Bad statetransition of testapp {site} during test to {state}"))
        self.error_message = ''

        settings = self.user_settings.copy()

        import copy
        self.testing_sites = settings['sites'] if settings.get('sites') else copy.deepcopy(self.sites)

        if not set(self.testing_sites).issubset(set(self.sites)):
            self.on_error(f'not all testing sites "{self.testing_sites}" are in the configured site list "{self.sites}"')

        settings.update(self._extract_sites_information(param_data))
        self.connectionHandler.send_next_to_all_sites(settings)

        self.testing_event.set()

    def _extract_sites_information(self, parameters):
        try:
            sites_info = parameters['sites']
            sites_to_test = [site['siteid'] for site in sites_info]
            sites_to_test.sort()
            self.sites.sort()

            sites = []
            sites.extend(sites_to_test)
            sites.extend(self.sites)
            sites = list(set(sites))

            if len(self.sites) < len(sites):
                self.log.log_message(LogLevel.Error(), f'Master do not support site(s): {set(sites) - set(self.sites)}')
                self.on_error(f'"next command", handler requires "{sites}" but seems some are not configured: configured sites "{self.sites}"')

            return self._generate_sites_message(sites_to_test, sites_info)
        except Exception:
            return self._generate_default_sites_configuration()

    def _generate_default_sites_configuration(self):
        sites_info = []
        for index in range(len(self.sites)):
            sites_info.append({'siteid': str(index), 'partid': str(self.dummy_partid), 'binning': -1})
            self.dummy_partid += 1

        return self._generate_sites_message(self.sites, sites_info)

    @staticmethod
    def _generate_sites_message(sites, sites_info):
        return {'sites': sites, 'sites_info': sites_info}

    def on_unload_command_issued(self, param_data: dict):
        self.arm_timeout(UNLOAD_TIMEOUT, lambda: self.timeout("not all sites unloaded the testprogram"))
        self.pendingTransitionsControl = SequenceContainer([CONTROL_STATE_IDLE], self.configuredSites, lambda: self.all_siteunloads_complete(),
                                                           lambda site, state: self.on_error(f"Bad statetransition of control {site} during unload to {state}"))
        self.pendingTransitionsTest = SequenceContainer([TEST_STATE_TERMINATED], self.configuredSites, lambda: None, lambda site, state: None)
        self.error_message = ''
        self.connectionHandler.send_terminate_to_all_sites()
        self.handle_reset_command()
        self._first_part_tested = False

    def on_reset_received(self, param_data: dict):
        self.arm_timeout(UNLOAD_TIMEOUT, lambda: self.timeout("not all sites do report state idle after reset"))
        self.pendingTransitionsControl = SequenceContainer([CONTROL_STATE_IDLE], self.configuredSites, lambda: self.all_sites_detected(),
                                                           lambda site, state: self.on_unexpected_control_state(site, state))

        self.pendingTransitionsTest = None
        self.connectionHandler.send_reset_to_all_sites()

        # set a virtual control state to make sure that any response from control would be caught and handled
        for site in self.configuredSites:
            self.siteStates[site] = 'resetting'

    def on_allsiteunloadscomplete(self):
        self.disarm_timeout()

        self.received_sites_test_results.clear()
        self.loaded_lot_number = ''
        self._reset_execution_strategy()

    def on_allsitetestscomplete(self):
        self.testing_event.clear()
        self._send_test_results()
        self.test_results = []
        self.disarm_timeout()
        self.handle_reset()
        self._reset_execution_strategy()

    def _reset_execution_strategy(self):
        try:
            self.testing_strategy.reset_stages()
        except Exception as e:
            self.on_error(e)

    def on_all_sites_request_testing(self):
        try:
            released_sites = self.testing_strategy.handle_release(self.testing_sites)
            self.handle_release(released_sites)
        except Exception as e:
            self.on_error(f'{self.on_all_sites_request_testing.__name__}: {e}')

    def _send_test_results(self):
        self.connectionHandler.send_test_results(self.test_results)

    def on_site_test_result_received(self, site_id, param_data):
        self.test_num = 0
        payload = param_data['payload']
        self._write_stdf_data(payload)

        prr_record = None
        # hack: (-)inf could not be parsed into a json object, so we cast it to string
        for index, rec in enumerate(payload):
            for key, value in rec.items():
                if str(value) in ('inf', '-inf'):
                    payload[index][key] = str(value)

            if rec['type'] == 'PRR':
                prr_record = rec

        assert prr_record

        self.received_sites_test_results.append(payload)
        self.received_site_test_results.append(param_data)
        success, msg = self._result_info_handler.handle_result(prr_record)

        if not success:
            self.on_error(msg)

        self.test_results.append(self._result_info_handler.get_site_result_response(prr_record))

    def _write_stdf_data(self, stdf_data):
        self._stdf_aggregator.append_test_results(stdf_data)

    def on_control_status_changed(self, siteid: str, status_msg: dict):
        if self.external_state == 'softerror':
            return

        newstatus = status_msg['payload']['state']

        if(status_msg['interface_version'] != INTERFACE_VERSION):
            self.log.log_message(LogLevel.Error(), f'Bad interface version on site {siteid}')
            self.bad_interface_version()

        try:
            if(self.siteStates[siteid] != newstatus):
                self.log.log_message(LogLevel.Info(), f'Control {siteid} state is {newstatus}')
                self.siteStates[siteid] = newstatus
                self.pendingTransitionsControl.trigger_transition(siteid, newstatus)
        except KeyError:
            self.on_error(f"Site id received: {siteid} is not configured")

    def on_testapp_status_changed(self, siteid: str, status_msg: dict):
        if self.external_state == 'softerror':
            return

        newstatus = status_msg['payload']['state']
        self.log.log_message(LogLevel.Info(), f'Testapp {siteid} state is {newstatus}')

        if self.pendingTransitionsTest:
            self.pendingTransitionsTest.trigger_transition(siteid, newstatus)

    def on_log_message(self, siteid: str, log_msg: dict):
        self.log.append_log(log_msg['payload'])

    def on_testapp_testresult_changed(self, siteid: str, status_msg: dict):
        if self.is_testing(allow_substates=True):
            self.on_site_test_result_received(siteid, status_msg)
            self.handle_testresult(siteid, status_msg)
        else:
            self.on_error(f"Received unexpected testresult from site {siteid}")

    def on_execution_strategy_message(self, siteid: str, execution_strategy: dict):
        if self.handle_execution_strategy_message(siteid, execution_strategy):
            self.set_execution_strategy_configuration(execution_strategy)

    def on_testapp_testsummary_changed(self, message: dict):
        self.summary_counter += 1
        self.tsr_messages.extend(message['payload'])

        if self.summary_counter == len(self.configuredSites):
            self._stdf_aggregator.finalize()

            self._stdf_aggregator.append_soft_and_hard_bin_record(self._result_info_handler.get_hbin_soft_bin_report())
            self._stdf_aggregator.append_test_summary(self.tsr_messages)
            self._stdf_aggregator.append_part_count_infos(self._result_info_handler.get_part_count_infos())

            self._stdf_aggregator.write_footer_records()
            self._stdf_aggregator = None
            self.summary_counter = 0
            self.tsr_messages.clear()

    def on_testapp_resource_changed(self, siteid: str, resource_request_msg: dict):
        self.handle_resource_request(siteid, resource_request_msg)

    def on_testapp_test_request_changed(self, siteid: str):
        self.handle_test_request(siteid)

    def on_peripheral_response_received(self, response):
        self.peripheral_controller.on_response_received(response)

    def on_handler_status_changed(self, msg: dict):
        if self.external_state == 'softerror':
            return

        if msg['state'] == 'connecting':
            self.connectionHandler.publish_state(self.state)

        # TODO: how to handle status crash or error
        if msg['state'] in ('error', 'crash'):
            return

    def on_handler_response_message(self, msg):
        res = msg.get('type')
        payload = msg.get('payload')
        try:
            {
                'temperature': lambda: self._handle_handler_temperature(payload),
            }[res]()
        except KeyError:
            self.log.log_message(LogLevel.Error(), f'Failed to execute response: {res}')
        except MachineError:
            self.on_error(f'cannot trigger command "{res}" from state "{self.fsm.model.state}"')

    def _handle_handler_temperature(self, param_data):
        self.log.log_message(LogLevel.Info(), f'handler temperature is : {param_data["temperature"]}')

    def on_handler_command_message(self, msg: dict):
        cmd = msg.get('type')
        payload = msg.get('payload')
        try:
            {
                'load': lambda param_data: self.load_command(param_data),
                'next': lambda param_data: self.next(param_data),
                'unload': lambda param_data: self.unload(param_data),
                'reset': lambda param_data: self.reset(param_data),
                'identify': lambda param_data: self._send_tester_identification(param_data),
                'get-state': lambda param_data: self._send_tester_state(param_data),
                'site-layout': lambda param_data: self._set_layout(param_data['sites']),
                'get-host': lambda _: self._send_host_link()
            }[cmd](payload)
        except KeyError:
            self.log.log_message(LogLevel.Error(), f'Failed to execute command: {cmd}')
        except MachineError:
            self.on_error(f'cannot trigger command "{cmd}" from state "{self.fsm.model.state}"')

    def _send_host_link(self):
        port = self.configuration['webui_port']
        host = self.configuration['webui_host']

        self.connectionHandler.send_host_info(host, port)

    def _set_layout(self, layout: list):
        if len(self.sites) < len(layout):
            self.on_error('number of sites use to be configured is less that configured,'
                          + f' handler interface provides "{len(layout)}" sites layout'
                          + f' but master is only configured to support "{len(self.sites)}" sites')
        self.layout = layout

    def _send_tester_identification(self, _):
        self.connectionHandler.send_identification()

    def _send_tester_state(self, _):
        self.connectionHandler.send_state(self.external_state, self.error_message)

    def send_handler_get_temperature_command(self):
        self.connectionHandler.send_handler_get_temperature_command()

    def apply_resource_config(self, resource_request: dict, on_resource_config_applied_callback: Callable):
        resource_id = resource_request['periphery_type']

        async def periphery_io_control_task():
            try:
                result = await self.periphery_io_control(resource_request)
                on_resource_config_applied_callback()
                self.connectionHandler.publish_ioctl_response(resource_request, result)
            except asyncio.TimeoutError:
                # Applying the resource configuration ran into a timeout.
                # We err-out in this case.
                self.connectionHandler.publish_ioctl_timeout(resource_request)
                # TBD: Dunno if this is actually that much of a good idea,
                # as nothing is wrong with the tester per se here.
                self.on_error(f"Failed to control resource {resource_id}. Reason: Timeout.")

        asyncio.get_event_loop().create_task(periphery_io_control_task())

    async def periphery_io_control(self, resource_request):
        resource_id = resource_request['periphery_type']
        ioctl_name = resource_request['ioctl_name']
        parameters = resource_request['parameters']
        return await self.peripheral_controller.device_io_control(resource_id, ioctl_name, parameters)

    def dispatch_command(self, json_data):
        cmd = json_data.get('command')
        try:
            {
                'load': lambda param_data: self.load_command(param_data),
                'next': lambda param_data: self.next(param_data),
                'unload': lambda param_data: self.unload(param_data),
                'reset': lambda param_data: self.reset(param_data),
                'usersettings': lambda param_data: self.usersettings_command(param_data),
                'getresults': lambda param_data: self._handle_command_with_response(param_data),
                'getlogs': lambda param_data: self._handle_command_with_response(param_data),
                'getlogfile': lambda param_data: self._handle_command_with_response(param_data),
                'setloglevel': lambda param_data: self.setloglevel(param_data),
                'binMap': lambda param_data: self._update_hbin_number(param_data),
            }[cmd](json_data)
        except Exception as e:
            self.log.log_message(LogLevel.Error(), f'Failed to execute command {cmd}: {e}')

    def _update_hbin_number(self, param_data: dict):
        # TODO: UI message should contain sbin number
        sbin = str(param_data['sBin'])
        hbin = str(param_data['hBin'])

        self._result_info_handler.update_hbin_number(sbin, hbin)
        self.connectionHandler.send_set_new_hbin(sbin, hbin)

    def on_setloglevel_command(self, param_data):
        loglevel = param_data['level']
        self.loglevel = loglevel
        self.log.set_logger_level(loglevel)
        self._send_set_log_level()
        self.command_queue.put_nowait(GetUserSettings(lambda: self._generate_usersettings_message(self.user_settings)))

    def _handle_command_with_response(self, data):
        cmd = data.get('command')
        connection_id = data.get('connectionid')
        try:
            obj = {'getlogs': lambda: GetLogsCommand(self.log, connection_id),
                   'getlogfile': lambda: GetLogFileCommand(self.log, connection_id),
                   'getresults': lambda: GetTestResultsCommand(self.received_sites_test_results, connection_id),
                   }[cmd]()
            try:
                self.command_queue.put_nowait(obj)
            except Full:
                pass

        except KeyError:
            pass

    def on_new_connection(self, connection_id):
        self.command_queue.put_nowait(GetUserSettings(lambda: self._generate_usersettings_message(self.user_settings)))
        self.command_queue.put_nowait(GetTestResultsCommand(self.received_sites_test_results, connection_id))
        self.command_queue.put_nowait(GetYields(lambda: self._generate_yield_message(), connection_id))
        self.command_queue.put_nowait(GetBinTable(lambda: self._generate_bin_table_message(), connection_id))

        if self._stdf_aggregator:
            self.command_queue.put_nowait(GetLotData(lambda: self._generate_lot_data_message(), connection_id))

    async def _execute_commands(self, ws_comm_handler):
        while True:
            try:
                command = self.command_queue.get_nowait()
            except Empty:
                break

            if not command.is_data_ready():
                try:
                    self.command_queue.put_nowait(command)
                except Full:
                    continue

            await command.execute(ws_comm_handler)

    async def _handle_sending_data(self, ws_comm_handler):
        # TODO: used to update front-end's clock, do we need to do this anyway
        await ws_comm_handler.send_status_to_all(self.external_state, self.error_message)

        # TODO: ATE-227, sync with UI Team
        for test_result in self.received_site_test_results:
            await ws_comm_handler.send_testresults_to_all(test_result)
            await ws_comm_handler.send_yields(self._generate_yield_message())
            await ws_comm_handler.send_bin_table(self._generate_bin_table_message())

        self.received_site_test_results = []

        await self._execute_commands(ws_comm_handler)

        available_logs = self.log.get_current_logs()
        if len(available_logs) > 0:
            await ws_comm_handler.send_logs_to_all(available_logs)

    def _generate_yield_message(self):
        return self._result_info_handler.get_yield_messages()

    def _generate_bin_table_message(self):
        return self._result_info_handler.get_bin_table()

    def _generate_usersettings_message(self, usersettings):
        settings = []

        for usersetting, value in usersettings.items():
            settings.append({'name': usersetting, 'active': value['active'], 'value': int(value['value'])})

        test_options = {'testoptions': settings}
        log_level = {'loglevel': self.loglevel}
        usersetting = {}
        usersetting.update(test_options)
        usersetting.update(log_level)

        return usersetting

    def _generate_lot_data_message(self):
        return self._stdf_aggregator.get_MIR_dict()

    async def _master_background_task(self, app):
        try:
            while True:
                ws_comm_handler = app['ws_comm_handler']
                if ws_comm_handler is None:
                    await asyncio.sleep(1)
                    continue

                if not ws_comm_handler.is_ws_connection_established():
                    self.log.clear_logs()

                await self._handle_sending_data(ws_comm_handler)
                await asyncio.sleep(1)

        except asyncio.CancelledError:
            pass

    async def _master_request_task(self, app):
        try:
            while True:
                await self.testing_event.wait()

                sites = await self.testing_strategy.get_site_states(1)
                self.handle_sites_request(sites)

        except asyncio.CancelledError:
            pass

    def run(self):
        try:
            task_handler = TaskHandler(self,
                                       self.configuration,
                                       self.connectionHandler,
                                       lambda app: self._master_request_task(app),
                                       lambda app: self._master_background_task(app))
            task_handler.run()
        except KeyboardInterrupt:
            pass



from ate_master_app.master_webservice import WebSocketConnection, WebsocketCommunicationHandler
import threading


class NonBlockingCall(threading.Thread):
    def __init__(self, callback):
        super().__init__(target=callback)
        self.start()


class NonBlockingCommand:
    def __init__(self, obj, connection_id=None):
        self._is_ready = False
        self.connection_id = connection_id
        self._reply = None
        NonBlockingCall(lambda: self.acquire_data(obj))

    def acquire_data(self, obj):
        self.acquire_data_impl(obj)
        self._is_ready = True

    def is_data_ready(self):
        return self._is_ready

    async def execute(self, ws_conn_handler: WebSocketConnection):
        if not self._reply:
            # skip executing if rely not read yet
            return

        await self.execute_impl(ws_conn_handler)

    async def execute_impl(self, ws_conn_handler: WebSocketConnection):
        raise Exception("Implement me")

    def acquire_data_impl(self, data):
        raise Exception("Implement me")


class GetLogFileCommand(NonBlockingCommand):
    async def execute_impl(self, ws_comm_handler):
        await ws_comm_handler.send_logfile(self._reply, self.connection_id)

    def acquire_data_impl(self, obj):
        self._reply = obj.get_log_file_information()


class GetTestResultsCommand(NonBlockingCommand):
    async def execute_impl(self, ws_comm_handler):
        await ws_comm_handler.send_testresults_to_client(self._reply, self.connection_id)

    def acquire_data_impl(self, data):
        self._reply = list(data.get_data())


class GetLogsCommand(NonBlockingCommand):
    async def execute_impl(self, ws_comm_handler):
        await ws_comm_handler.send_logs(self._reply, self.connection_id)

    def acquire_data_impl(self, data):
        self._reply = data.get_logs()


class GetUserSettings(NonBlockingCommand):
    async def execute_impl(self, ws_comm_handler):
        await ws_comm_handler.send_user_settings(self._reply)

    def acquire_data_impl(self, get_user_settings_call_back):
        self._reply = get_user_settings_call_back()


class GetYields(NonBlockingCommand):
    async def execute_impl(self, ws_comm_handler):
        await ws_comm_handler.send_yields(self._reply)

    def acquire_data_impl(self, get_yield_call_back):
        self._reply = get_yield_call_back()


class GetLotData(NonBlockingCommand):
    async def execute_impl(self, ws_comm_handler: WebsocketCommunicationHandler):
        await ws_comm_handler.send_lotdata(self._reply)

    def acquire_data_impl(self, get_lotdata_call_back: callable):
        self._reply = get_lotdata_call_back()


class GetBinTable(NonBlockingCommand):
    async def execute_impl(self, ws_comm_handler: WebsocketCommunicationHandler):
        await ws_comm_handler.send_bin_table(self._reply)

    def acquire_data_impl(self, get_lotdata_call_back: callable):
        self._reply = get_lotdata_call_back()



from setuptools import find_packages, setup
from pathlib import Path
from ate_master_app import __version__

version = __version__
requirements_path = Path(Path(__file__).parents[0], 'requirements/run.txt')
def add_version(name: str) -> str:
    return f'{name.rstrip()}=={version}' if 'ate' in name else name.rstrip()
       
with requirements_path.open('r') as f:
    install_requires = list(map(add_version, f))

readme_path = Path(Path(__file__).parent, 'ate_master_app/README.md')
with readme_path.open('r') as f:
    long_description = f.read()

setup(
    name='semi-ate-master-app',
    version=version,
    description='Master application used for steering the control applications in an ATE test environment.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="The Semi-ATE Project Contributors",
    author_email="ate.organization@gmail.com",
    license="GPL2",
    keywords="Semiconductor ATE Automatic Test Equipment Spyder Plugin",
    platforms=["Windows", "Linux", "Mac OS-X"],
    packages=find_packages(),
    install_requires=install_requires,
    include_package_data=True,
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance',
    ],
    entry_points={
        'console_scripts': ['launch_master=ate_master_app.launch_master:main']
    },
)



from transitions.extensions import HierarchicalMachine as Machine

from ate_master_app.statemachines.TestingSiteMachine import (TestingSiteModel, TestingSiteMachine)

from typing import List


class MultiSiteTestingMachine(Machine):

    def __init__(self, model=None):
        states = ['inprogress', 'waiting_for_resource', 'completed', 'wating_for_release']
        super().__init__(model=model, states=states, initial='inprogress')

        self.add_transition('all_sites_waiting_for_resource',       '*',                                'waiting_for_resource')      # noqa: E241
        self.add_transition('resource_config_applied',              'waiting_for_resource',                      'inprogress')                # noqa: E241
        self.add_transition('all_sites_completed',                  '*',                                         'completed')                 # noqa: E241
        self.add_transition('all_sites_request_testing',            '*',                                         'inprogress')                # noqa: E241


class MultiSiteTestingModel:
    def __init__(self, site_ids: List[str], parent_model=None):
        self._site_models = {site_id: TestingSiteModel(site_id) for site_id in site_ids}
        self._site_machines = {site_id: TestingSiteMachine(self._site_models[site_id]) for site_id in site_ids}
        self._parent_model = parent_model if parent_model is not None else self

        self._site_test_result_received = {site_id: False for site_id in site_ids}
        self._released_sites = []

    def handle_reset(self):
        for site in self._site_models.values():
            if site.is_completed():
                site.reset()

        self.handle_reset_command()

    def handle_resource_request(self, site_id: str, resource_request: dict):
        self._site_models[site_id].resource_requested(resource_request=resource_request)

        for site in self._site_models.values():
            if site.resource_request is not None and site.resource_request != resource_request:
                raise RuntimeError(f'mismatch in resource request from site "{site_id}": previous request of site "{site.site_id}" differs')

        self._handle_resource_release()

    def _on_resource_config_applied(self):
        self.resource_config_applied()
        for site in self._site_models.values():
            if site.is_testing_waiting_for_resource():
                site.resource_ready()

    def _handle_resource_release(self):
        if not self._check_for_all_remaining_sites_waiting_for_resource():
            return

        resource_request = self._site_models[self._released_sites[0]].resource_request  # all sites have same request
        self._resource_required(resource_request)

    def handle_testresult(self, site_id: str, testresult: dict):
        self._site_test_result_received[site_id] = True

        self._site_models[site_id].testrequest_received()
        self._released_sites.pop(self._released_sites.index(site_id))
        self._handle_resource_release()

        if not self._check_for_all_sites_completed():
            return

        self.all_sites_completed()
        self._parent_model.all_sitetests_complete()
        self._reset_sites_result_received()

    def _reset_sites_result_received(self):
        for site_id in self._site_test_result_received:
            self._site_test_result_received[site_id] = False

    def handle_status_idle(self, site_id: str):
        # TODO: do we need this
        return

    def handle_sites_request(self, sites: List[int]):
        for site_id in range(len(sites)):
            site_id = str(site_id)
            if not self._site_models.get(str(site_id)):
                continue

            if self._site_models[site_id].is_testing_idle() and sites[int(site_id)]:
                self._site_models[site_id].testrequest_received()

            if self._site_models[site_id].is_testing_waiting_for_release() and not sites[int(site_id)]:
                self._site_models[site_id].testrequest_released()

            if self._site_models[site_id].is_testing_busy() and sites[int(site_id)]:
                self._site_models[site_id].testrequest_received()

        self.handle_test_request(str(site_id))

    def handle_test_request(self, site_id: str):
        if all(site.is_testing_waiting_for_release() for site in self._site_models.values()) and self.is_testing_inprogress():
            self.all_sites_request_testing()
            self._parent_model.all_site_request_testing()

    def handle_release(self, sites):
        self._released_sites = sites
        for site_id in sites:
            self._site_models[site_id].testrequest_released()

    def _check_for_all_sites_completed(self):
        return all(is_received for _, is_received in self._site_test_result_received.items())

    def _check_for_all_remaining_sites_waiting_for_resource(self):
        sites_waiting = [site_id for site_id, site in self._site_models.items() if site.is_testing_waiting_for_resource()]
        if set(sites_waiting) != set(self._released_sites) or not len(sites_waiting):
            return False

        return True

    def _resource_required(self, resource_request: dict):
        self.all_sites_waiting_for_resource()
        self._parent_model.apply_resource_config(resource_request, lambda: self._on_resource_config_applied())  # Callable[[dict, Callable], None]

    def handle_reset_command(self):
        for site in self._site_models.values():
            site.on_unload()

    def handle_execution_strategy_message(self, site_id: str, execution_strategy: List[List[List[str]]]):
        self._site_models[site_id].set_execution_strategy(execution_strategy)

        for site in self._site_models.values():
            if site.execution_strategy is not None and site.execution_strategy != execution_strategy:
                raise RuntimeError(f'mismatch in execution strategy configuration from site "{site_id}": "{execution_strategy}"\
                                   previous strategy configuration of site "{site.site_id}": "{site.execution_strategy}" differs')

        if (all([site.execution_strategy is not None for site in self._site_models.values()])):
            return True

        return False



from ate_common.program_utils import BinTableFieldName
from ate_common.logger import LogLevel
import os
import re


class FileSystemDataSource:
    def __init__(self, jobname, configuration, parser, logger):
        self.parser = parser
        self.configuration = configuration.get('jobformat')
        self.device_id = configuration.get('device_id')
        self.device_handler = configuration.get("Handler")
        self.device_environment = configuration.get("Environment")

        self.path = configuration['filesystemdatasource.path']
        # the jobpattern defines a pattern for
        # the filename, e.g. le#jobname#.xml
        self.lot_number = jobname
        self.jobname = configuration['filesystemdatasource.jobpattern'].\
            replace("#jobname#", jobname)

        self.log = logger
        self.fullpath = os.path.join(self.path, self.jobname)

    def does_file_exist(self):
        if not os.path.exists(self.fullpath):
            self.log.log_message(LogLevel.Error(), f'job file "{self.fullpath}" could not be not found')
            return False

        return True

    def retrieve_data(self):
        if not self.does_file_exist():
            return None

        if self.configuration is None:
            return None

        if self.parser is None:
            return None

        param_data = self.parser.parse(self.fullpath)
        for key, value in param_data.items():
            if value is None:
                self.log.log_message(LogLevel.Error(), f'{key} section is missing')
                return None

        return param_data

    def verify_data(self, param_data):
        return (self.is_lot_number_valid(param_data.get("MAIN"), self.lot_number)
                and self.does_device_id_exist(param_data.get("CLUSTER"))
                and self.is_station_valid(param_data.get("STATION"))
                and self.does_handler_id_exist(param_data.get("CLUSTER")))

    def get_test_information(self, param_data: dict):
        test_information = {}
        test_information['PACKAGE_ID'] = param_data['MAIN']['PACKAGEID_SHORT']
        test_information['SUBLOT_ID'] = param_data['MAIN']['LOTNUMBER'].split('.')[1]
        for _, value in param_data.get("STATION").items():
            if self.does_device_id_exist(value):
                test_information.update(self.remove_digit_from_keys(value))
                test_information['PART_ID'] = f"{param_data['MAIN']['MATCHCODE']}_{test_information['TESTERPRG']}"
                return test_information

        self.log.log_message(LogLevel.Warning(), f'device information for "{self.device_id}" in the STATION section are missing')
        return None

    def remove_digit_from_keys(self, data: dict):
        raw_data = {}
        for key, value in data.items():
            raw_data[''.join(i for i in key if not i.isdigit())] = value

        return raw_data

    def is_lot_number_valid(self, param_data: dict, lot_number):
        if not (param_data.get("LOTNUMBER") == lot_number):
            self.log.log_message(LogLevel.Warning(),
                                 f'lot number: {lot_number} could not be found, lot number defined in xml file: {param_data.get("LOTNUMBER")}')
            return False

        return True

    def does_handler_id_exist(self, param_data: dict):
        if not (param_data.get("HANDLER_PROBER") == self.device_handler):
            self.log.log_message(LogLevel.Warning(),
                                 f'HANDLER: {self.device_handler} name could not be found, handler name defined in xml file: {param_data.get("HANDLER_PROBER")}')
            return False

        return True

    def does_device_id_exist(self, param_data: dict):
        for key, value in param_data.items():
            if key == 'TESTER' or re.match(r'TESTER\d', key) is not None:
                if self.device_id == value:
                    return True

        return False

    def is_station_valid(self, param_data: dict):
        for _, value in param_data.items():
            if self.does_device_id_exist(value):
                return True

        self.log.log_message(LogLevel.Warning(), f'device id {self.device_id} could not be found in STATION section')
        return False

    @staticmethod
    def get_bin_table(data: list):
        bin_table = []
        for entry in data['BINTABLE']['ENTRY']:
            bin_element = {}
            for key, value in entry.items():
                key = key.replace('@', '')
                bin_element[key] = value

            bin_table.append(bin_element)

        return bin_table

    @staticmethod
    def get_binning_tuple(bin_table: list) -> dict:
        binning_tuple = {}
        for bin_info in bin_table:
            binning_tuple.setdefault(bin_info[BinTableFieldName.HBin()], []).append(bin_info[BinTableFieldName.SBinNum()])

        return binning_tuple



from aiohttp import web
import asyncio
from enum import Enum
from ate_master_app.master_connection_handler import MasterConnectionHandler
from ate_master_app.master_webservice import webservice_setup_app

class TaskType(Enum):
    Websocket = 'websocket'
    Request = 'request'
    Mqtt = 'mqtt'

    def __call__(self):
        return self.value


class TaskHandler:
    def __init__(self, parent: object, configuration: dict, connection_handler: MasterConnectionHandler, request_callback: callable, websocket_callback: callable):
        self._connection_handler = connection_handler
        self._request_handler = request_callback
        self._websocket_handler = websocket_callback
        self._configuration = configuration
        self.parent = parent
        self.app = web.Application()
        self.app['master_app'] = self.parent

    @staticmethod
    def _create_task_callback(callback: callable) -> callable:
        async def task_ctx(app):
            task = asyncio.create_task(callback(app))

            yield

            task.cancel()
            await task

        return lambda app: task_ctx(app)

    @staticmethod
    def _run_task_callback(callback: callable) -> callable:
        async def task_ctx(app):
            callback.start()

            yield

            await callback.stop()

        return lambda app: task_ctx(app)

    def run(self):
        web_root_path = self._configuration.get('webui_root_path')
        webservice_setup_app(self.app, web_root_path)
        self.app.cleanup_ctx.append(self._run_task_callback(self._connection_handler))
        self.app.cleanup_ctx.append(self._create_task_callback(lambda app: self._websocket_handler(app)))
        self.app.cleanup_ctx.append(self._create_task_callback(lambda app: self._request_handler(app)))
        host = self._configuration.get('webui_host', 'localhost')
        port = self._configuration.get('webui_port', 8081)
        web.run_app(self.app, host=host, port=port)



from ate_common.program_utils import BinTableFieldName


class YieldInformation:
    def __init__(self, types, site_id):
        self._site_id = site_id
        self._yield_info = {}
        self._init_yield_info(types)

    def _init_yield_info(self, types):
        for t in types:
            self._yield_info[t] = {"count": 0, "value": 0}

    def update_yield_info(self, type):
        self._yield_info[type]["count"] += 1
        self._calculate_yield()

    def max_count(self):
        return sum([values["count"] for _, values in self._yield_info.items()])

    def get_yield_info(self):
        return self._yield_info

    def _calculate_yield(self):
        max_count = self.max_count()
        for _, t in self._yield_info.items():
            t["value"] = self._calculate_value(t["count"], max_count)

    @staticmethod
    def _calculate_value(count, max):
        return (count / max) * 100

    def generate_yield_messages(self):
        messages = [self._yield_message(name, values["value"], values["count"])
                    for name, values in self._yield_info.items()]

        messages.append(self._yield_message('sum', 100.0, self.max_count()))
        return messages

    def _yield_message(self, name, value, count):
        return {"name": name, "siteid": self._site_id, "value": value, "count": count}


class YieldInformationHandler:
    def __init__(self):
        self._bin_settings = None
        self.prr_rec_information = {}
        self.sites_yield_information = {}

    @staticmethod
    def _get_part_id(prr_record):
        part_id = ''
        if prr_record['PART_ID']:
            part_id = prr_record['PART_ID']
        else:
            default_value = -32768
            if (prr_record['X_COORD'], prr_record['Y_COORD']) == (default_value, default_value):
                assert False
            else:
                part_id = f"{prr_record['X_COORD']}x{prr_record['Y_COORD']}"

        return part_id

    def get_site_result(self, prr_record):
        part_id = self._get_part_id(prr_record)
        hard_bin = prr_record['HARD_BIN']
        siteid = str(prr_record['SITE_NUM'])

        return {'siteid': siteid, 'partid': part_id, 'binning': hard_bin, 'logflag': 0, 'additionalinfo': 0}

    def accumulate_site_bin_info(self, site_num, soft_bin):
        if not self._bin_settings:
            return False, "bin settings are not received yet"

        for typ, setting in self._bin_settings.items():
            if soft_bin not in setting['sbins']:
                continue

            if not self.sites_yield_information.get(site_num):
                self.sites_yield_information[site_num] = YieldInformation(list(self._bin_settings.keys()), site_num)

            self.sites_yield_information[site_num].update_yield_info(typ)
            return self._accumulate_all_bin_info()

        return False, f"soft bin: '{soft_bin}' could not be mapped to any of the setting's bins"

    def _accumulate_all_bin_info(self):
        all_id = '-1'
        all_yield_info = YieldInformation(list(self._bin_settings.keys()), all_id)
        for site_id, info in self.sites_yield_information.items():
            if site_id == all_id:
                continue

            for typ, values in info.get_yield_info().items():
                all_yield_info._yield_info[typ]["count"] += values["count"]

        all_yield_info._calculate_yield()
        self.sites_yield_information[all_id] = all_yield_info
        return True, ''

    def reaccumulate_bin_info(self, prr_rec_information):
        self.sites_yield_information.clear()

        if len(prr_rec_information) == 0:
            return self._accumulate_all_bin_info()

        for _, prr_rec in prr_rec_information.items():
            site_num = str(prr_rec['SITE_NUM'])
            soft_bin = str(prr_rec['SOFT_BIN'])
            result = self.accumulate_site_bin_info(site_num, soft_bin)
            if not result[0]:
                return result

        return True, ''

    def set_bin_settings(self, bin_settings: dict):
        self._bin_settings = bin_settings

    @staticmethod
    def get_bin_settings(bin_table: list) -> dict:
        settings = {}

        for bin_info in bin_table:
            settings.setdefault(bin_info[BinTableFieldName.SBinGroup()], {'hbins': [], 'sbins': []})['hbins'].append(bin_info[BinTableFieldName.HBin()])
            settings.setdefault(bin_info[BinTableFieldName.SBinGroup()], {'hbins': [], 'sbins': []})['sbins'].append(bin_info[BinTableFieldName.SBinNum()])

        return settings

    def get_yield_messages(self):
        messages = []
        for _, yield_info in self.sites_yield_information.items():
            messages.extend(yield_info.generate_yield_messages())

        return messages

    def get_site_yield_info_message(self, site: str):
        return self.sites_yield_information[site].generate_yield_messages()

    def clear_yield_information(self):
        self.sites_yield_information.clear()
